package com.sun.jndi.ldap;

import java.util.Arrays;
import sun.misc.IOUtils;
import java.io.InterruptedIOException;
import javax.naming.ldap.Control;
import javax.naming.InterruptedNamingException;
import javax.naming.ServiceUnavailableException;
import javax.net.ssl.SSLParameters;
import java.lang.reflect.Method;
import java.lang.reflect.Constructor;
import javax.net.ssl.SSLSocket;
import javax.naming.NamingException;
import java.lang.reflect.InvocationTargetException;
import javax.naming.CommunicationException;
import java.io.BufferedOutputStream;
import java.io.BufferedInputStream;
import java.security.AccessController;
import java.io.IOException;
import java.net.Socket;
import java.io.InputStream;
import java.io.OutputStream;

public final class Connection implements Runnable
{
    private static final boolean debug = false;
    private static final int dump = 0;
    private final Thread worker;
    private boolean v3;
    public final String host;
    public final int port;
    private boolean bound;
    private OutputStream traceFile;
    private String traceTagIn;
    private String traceTagOut;
    public InputStream inStream;
    public OutputStream outStream;
    public Socket sock;
    private final LdapClient parent;
    private int outMsgId;
    private LdapRequest pendingRequests;
    volatile IOException closureReason;
    volatile boolean useable;
    int readTimeout;
    int connectTimeout;
    private volatile boolean isUpgradedToStartTls;
    final Object startTlsLock;
    private static final boolean IS_HOSTNAME_VERIFICATION_DISABLED;
    private final Object pauseLock;
    private boolean paused;
    
    private static boolean hostnameVerificationDisabledValue() {
        String s = (String)AccessController.doPrivileged(()/*  */ -> System.getProperty("com.sun.jndi.ldap.object.disableEndpointIdentification"));
        return s != null && (s.isEmpty() || Boolean.parseBoolean(s));
    }
    
    void setV3(boolean v3) {
        this.v3 = v3;
    }
    
    void setBound() {
        this.bound = true;
    }
    
    Connection(LdapClient parent, String str, int i, String s, int connectTimeout, int readTimeout, OutputStream traceFile) throws NamingException {
        this.v3 = true;
        this.bound = false;
        this.traceFile = null;
        this.traceTagIn = null;
        this.traceTagOut = null;
        this.outMsgId = 0;
        this.pendingRequests = null;
        this.closureReason = null;
        this.useable = true;
        this.startTlsLock = new Object();
        this.pauseLock = new Object();
        this.paused = false;
        this.host = str;
        this.port = i;
        this.parent = parent;
        this.readTimeout = readTimeout;
        this.connectTimeout = connectTimeout;
        if (traceFile != null) {
            this.traceFile = traceFile;
            this.traceTagIn = "<- " + str + ":" + i + "\n\n";
            this.traceTagOut = "-> " + str + ":" + i + "\n\n";
        }
        try {
            this.sock = this.createSocket(str, i, s, connectTimeout);
            this.inStream = new BufferedInputStream(this.sock.getInputStream());
            this.outStream = new BufferedOutputStream(this.sock.getOutputStream());
        }
        catch (InvocationTargetException ex) {
            Throwable targetException = ex.getTargetException();
            CommunicationException ex2 = new CommunicationException(str + ":" + i);
            ex2.setRootCause(targetException);
            throw ex2;
        }
        catch (Exception rootCause) {
            CommunicationException ex3 = new CommunicationException(str + ":" + i);
            ex3.setRootCause(rootCause);
            throw ex3;
        }
        (this.worker = Obj.helper.createThread(this)).setDaemon(true);
        this.worker.start();
    }
    
    private Object createInetSocketAddress(String s, int value) throws NoSuchMethodException {
        try {
            return Class.forName("java.net.InetSocketAddress").getConstructor(new Class[] { String.class, Integer.TYPE }).newInstance(new Object[] { s, new Integer(value) });
        }
        catch (ClassNotFoundException | InstantiationException | InvocationTargetException | IllegalAccessException ex) {
            throw new NoSuchMethodException();
        }
    }
    
    private Socket createSocket(String host, int n, String s, int soTimeout) throws Exception {
        Socket socket = null;
        if (s != null) {
            Class<Object> loadClass = Obj.helper.loadClass(s);
            Object invoke = loadClass.getMethod("getDefault", new Class[0]).invoke(null, new Object[0]);
            if (soTimeout > 0) {
                try {
                    Method method2 = loadClass.getMethod("createSocket", new Class[0]);
                    Method method3 = Socket.class.getMethod("connect", new Class[] { Class.forName("java.net.SocketAddress"), Integer.TYPE });
                    Object inetSocketAddress2 = this.createInetSocketAddress(host, n);
                    socket = (Object)method2.invoke(invoke, new Object[0]);
                    method3.invoke(socket, new Object[] { inetSocketAddress2, new Integer(soTimeout) });
                }
                catch (NoSuchMethodException ex2) {}
            }
            if (socket == null) {
                socket = (Object)loadClass.getMethod("createSocket", new Class[] { String.class, Integer.TYPE }).invoke(invoke, new Object[] { host, new Integer(n) });
            }
        }
        else {
            if (soTimeout > 0) {
                try {
                    Constructor<Socket> constructor = Socket.class.getConstructor(new Class[0]);
                    Method method = Socket.class.getMethod("connect", new Class[] { Class.forName("java.net.SocketAddress"), Integer.TYPE });
                    Object inetSocketAddress = this.createInetSocketAddress(host, n);
                    socket = (Object)constructor.newInstance(new Object[0]);
                    method.invoke(socket, new Object[] { inetSocketAddress, new Integer(soTimeout) });
                }
                catch (NoSuchMethodException ex) {}
            }
            if (socket == null) {
                socket = new Socket(host, n);
            }
        }
        if ((socket instanceof SSLSocket)) {
            SSLSocket sslSocket = (SSLSocket)socket;
            if (!Connection.IS_HOSTNAME_VERIFICATION_DISABLED) {
                SSLParameters sslParameters = sslSocket.getSSLParameters();
                sslParameters.setEndpointIdentificationAlgorithm("LDAPS");
                sslSocket.setSSLParameters(sslParameters);
            }
            if (soTimeout > 0) {
                int soTimeout2 = sslSocket.getSoTimeout();
                sslSocket.setSoTimeout(soTimeout);
                sslSocket.startHandshake();
                sslSocket.setSoTimeout(soTimeout2);
            }
        }
        return socket;
    }
    
    synchronized int getMsgId() {
        return ++this.outMsgId;
    }
    
    LdapRequest writeRequest(BerEncoder berEncoder, int n) throws IOException {
        return this.writeRequest(berEncoder, n, false, -1);
    }
    
    LdapRequest writeRequest(BerEncoder berEncoder, int n, boolean b) throws IOException {
        return this.writeRequest(berEncoder, n, b, -1);
    }
    
    LdapRequest writeRequest(BerEncoder berEncoder, int n, boolean b, int n2) throws IOException {
        LdapRequest ldapRequest = new LdapRequest(n, b, n2);
        this.addRequest(ldapRequest);
        if (this.traceFile != null) {
            Ber.dumpBER(this.traceFile, this.traceTagOut, berEncoder.getBuf(), 0, berEncoder.getDataLen());
        }
        this.unpauseReader();
        try {
            synchronized (this) {
                this.outStream.write(berEncoder.getBuf(), 0, berEncoder.getDataLen());
                this.outStream.flush();
            }
        }
        catch (IOException closureReason) {
            this.cleanup(null, true);
            throw this.closureReason = closureReason;
        }
        return ldapRequest;
    }
    
    BerDecoder readReply(LdapRequest ldapRequest) throws IOException, NamingException {
        synchronized (this) {
            if (this.sock == null) {
                throw new ServiceUnavailableException(this.host + ":" + this.port + "; socket closed");
            }
        }
        NamingException ex = null;
        BerDecoder replyBer;
        try {
            replyBer = ldapRequest.getReplyBer((long)this.readTimeout);
        }
        catch (InterruptedException ex2) {
            throw new InterruptedNamingException("Interrupted during LDAP operation");
        }
        catch (CommunicationException ex3) {
            throw ex3;
        }
        catch (NamingException ex4) {
            ex = ex4;
            replyBer = null;
        }
        if (replyBer == null) {
            this.abandonRequest(ldapRequest, (Control[])null);
        }
        if (ex != null) {
            throw ex;
        }
        return replyBer;
    }
    
    private synchronized void addRequest(LdapRequest ldapRequest) {
        if (this.pendingRequests == null) {
            this.pendingRequests = ldapRequest;
            ldapRequest.next = null;
        }
        else {
            ldapRequest.next = this.pendingRequests;
            this.pendingRequests = ldapRequest;
        }
    }
    
    synchronized LdapRequest findRequest(int n) {
        LdapRequest ldapRequest = this.pendingRequests;
        while (ldapRequest != null) {
            if (ldapRequest.msgId == n) {
                return ldapRequest;
            }
            ldapRequest = ldapRequest.next;
        }
        return null;
    }
    
    synchronized void removeRequest(LdapRequest ldapRequest) {
        LdapRequest ldapRequest2 = this.pendingRequests;
        LdapRequest ldapRequest3 = null;
        while (ldapRequest2 != null) {
            if (ldapRequest2 == ldapRequest) {
                ldapRequest2.cancel();
                if (ldapRequest3 != null) {
                    ldapRequest3.next = ldapRequest2.next;
                }
                else {
                    this.pendingRequests = ldapRequest2.next;
                }
                ldapRequest2.next = null;
            }
            ldapRequest3 = ldapRequest2;
            ldapRequest2 = ldapRequest2.next;
        }
    }
    
    void abandonRequest(LdapRequest ldapRequest, Control[] array) {
        this.removeRequest(ldapRequest);
        BerEncoder berEncoder = new BerEncoder(256);
        int msgId = this.getMsgId();
        try {
            berEncoder.beginSeq(48);
            berEncoder.encodeInt(msgId);
            berEncoder.encodeInt(ldapRequest.msgId, 80);
            if (this.v3) {
                LdapClient.encodeControls(berEncoder, array);
            }
            berEncoder.endSeq();
            if (this.traceFile != null) {
                Ber.dumpBER(this.traceFile, this.traceTagOut, berEncoder.getBuf(), 0, berEncoder.getDataLen());
            }
            synchronized (this) {
                this.outStream.write(berEncoder.getBuf(), 0, berEncoder.getDataLen());
                this.outStream.flush();
            }
        }
        catch (IOException ex) {}
    }
    
    synchronized void abandonOutstandingReqs(Control[] array) {
        for (LdapRequest pendingRequests = this.pendingRequests; pendingRequests != null; pendingRequests = (this.pendingRequests = pendingRequests.next)) {
            this.abandonRequest(pendingRequests, array);
        }
    }
    
    private void ldapUnbind(Control[] array) {
        BerEncoder berEncoder = new BerEncoder(256);
        int msgId = this.getMsgId();
        try {
            berEncoder.beginSeq(48);
            berEncoder.encodeInt(msgId);
            berEncoder.encodeByte(66);
            berEncoder.encodeByte(0);
            if (this.v3) {
                LdapClient.encodeControls(berEncoder, array);
            }
            berEncoder.endSeq();
            if (this.traceFile != null) {
                Ber.dumpBER(this.traceFile, this.traceTagOut, berEncoder.getBuf(), 0, berEncoder.getDataLen());
            }
            synchronized (this) {
                this.outStream.write(berEncoder.getBuf(), 0, berEncoder.getDataLen());
                this.outStream.flush();
            }
        }
        catch (IOException ex) {}
    }
    
    void cleanup(Control[] array, boolean b) {
        boolean n = false;
        synchronized (this) {
            this.useable = false;
            if (this.sock != null) {
                try {
                    if (!b) {
                        this.abandonOutstandingReqs(array);
                    }
                    if (this.bound) {
                        this.ldapUnbind(array);
                    }
                    try {
                        this.outStream.flush();
                        this.sock.close();
                        this.unpauseReader();
                    }
                    catch (IOException ex) {}
                    if (!b) {
                        for (LdapRequest ldapRequest = this.pendingRequests; ldapRequest != null; ldapRequest = ldapRequest.next) {
                            ldapRequest.cancel();
                        }
                    }
                    this.sock = null;
                }
                finally {
                    try {
                        this.outStream.flush();
                        this.sock.close();
                        this.unpauseReader();
                    }
                    catch (IOException ex2) {}
                    if (!b) {
                        for (LdapRequest ldapRequest2 = this.pendingRequests; ldapRequest2 != null; ldapRequest2 = ldapRequest2.next) {
                            ldapRequest2.cancel();
                        }
                    }
                    this.sock = null;
                }
                n = b;
            }
            if (!(n == 0)) {
                for (LdapRequest ldapRequest3 = this.pendingRequests; ldapRequest3 != null; ldapRequest3 = ldapRequest3.next) {
                    ldapRequest3.close();
                }
            }
        }
        if (!(n == 0)) {
            this.parent.processConnectionClosure();
        }
    }
    
    public synchronized void replaceStreams(InputStream inStream, OutputStream outStream) {
        this.inStream = inStream;
        try {
            this.outStream.flush();
        }
        catch (IOException ex) {}
        this.outStream = outStream;
    }
    
    public synchronized void replaceStreams(InputStream inputStream, OutputStream outputStream, boolean isUpgradedToStartTls) {
        synchronized (this.startTlsLock) {
            this.replaceStreams(inputStream, outputStream);
            this.isUpgradedToStartTls = isUpgradedToStartTls;
        }
    }
    
    public boolean isUpgradedToStartTls() {
        return this.isUpgradedToStartTls;
    }
    
    private synchronized InputStream getInputStream() {
        return this.inStream;
    }
    
    private void unpauseReader() throws IOException {
        synchronized (this.pauseLock) {
            if (this.paused) {
                this.paused = false;
                this.pauseLock.notify();
            }
        }
    }
    
    private void pauseReader() throws IOException {
        this.paused = true;
        try {
            while (this.paused) {
                this.pauseLock.wait();
            }
        }
        catch (InterruptedException ex) {
            throw new InterruptedIOException("Pause/unpause reader has problems.");
        }
    }
    
    @Override
    public void run() {
        InputStream inputStream = null;
        try {
            while (true) {
                try {
                    byte[] array = new byte[129];
                    int n = 0;
                    int i = 0;
                    inputStream = this.getInputStream();
                    if (inputStream.read(array, n, 1) < 0) {
                        if (inputStream != this.getInputStream()) {
                            continue;
                        }
                        break;
                    }
                    else {
                        if (array[n++] != 48) {
                            continue;
                        }
                        int j = inputStream.read(array, n, 1);
                        if (j < 0) {
                            break;
                        }
                        int n2 = array[n++];
                        if ((n2 & 0x80) == 0x80) {
                            i = (n2 & 0x7F);
                            if (i > 4) {
                                throw new IOException("Length coded with too many bytes: " + i);
                            }
                            j = 0;
                            boolean n3 = false;
                            while (j < i) {
                                int read = inputStream.read(array, n + j, i - j);
                                if (read < 0) {
                                    n3 = true;
                                    break;
                                }
                                j += read;
                            }
                            if (!(n3 == 0)) {
                                break;
                            }
                            n2 = 0;
                            for (int k = 0; k < i; ++k) {
                                n2 = (n2 << 8) + (array[n + k] & 0xFF);
                            }
                            n += j;
                        }
                        if (i > j) {
                            throw new IOException("Unexpected EOF while reading length");
                        }
                        if (n2 < 0) {
                            throw new IOException("Length too big: " + ((long)n2 & 0xFFFFFFFFL));
                        }
                        byte[] fully = IOUtils.readFully(inputStream, n2, false);
                        byte[] copy = Arrays.copyOf(array, n + fully.length);
                        System.arraycopy(fully, 0, copy, n, fully.length);
                        int n4 = n + fully.length;
                        try {
                            BerDecoder o = new BerDecoder(copy, 0, n4);
                            if (this.traceFile != null) {
                                Ber.dumpBER(this.traceFile, this.traceTagIn, copy, 0, n4);
                            }
                            o.parseSeq((int[])null);
                            int int1 = o.parseInt();
                            o.reset();
                            if (int1 == 0) {
                                this.parent.processUnsolicited(o);
                            }
                            else {
                                LdapRequest request = this.findRequest(int1);
                                if (request != null) {
                                    synchronized (this.pauseLock) {
                                        if (request.addReplyBer(o)) {
                                            this.pauseReader();
                                        }
                                    }
                                }
                                else {
                                    continue;
                                }
                            }
                        }
                        catch (Ber.DecodeException ex) {}
                        continue;
                    }
                }
                catch (IOException ex2) {
                    if (inputStream != this.getInputStream()) {
                        continue;
                    }
                    throw ex2;
                }
            }
        }
        catch (IOException closureReason) {
            this.closureReason = closureReason;
        }
        finally {
            this.cleanup(null, true);
        }
    }
    
    static {
        IS_HOSTNAME_VERIFICATION_DISABLED = hostnameVerificationDisabledValue();
    }
}
